#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <ACG.h>

#include "standard.h"


struct Point_t
{
  unsigned int x;
  unsigned int y;
  Point_t(unsigned int xx=0, unsigned int yy=0) :x(xx), y(yy) {}

  void Print()
  {
	cout << '(' << x << ',' << y << ')' << endl ;
  }	

  operator == (const Point_t& aPoint)
  {
	return x == aPoint.x && y == aPoint.y ;
  }	

  operator < (const Point_t& aPoint)
  {
	return x < aPoint.x && y < aPoint.y ;
  }

  friend ostream&
  operator << (ostream& os, const Point_t& aPoint)
  {
	cout << '(' << aPoint.x << ',' << aPoint.y << ')' ;

	return os ; 
  }
};


/* Some constants used in data generation */

const unsigned extentp2x = 1<<24;
const unsigned extentp2y = 1<<24;
const int influence = 1024;
const int step = 20;
const int NCorners = 32;


extern "C" long random(void);
extern "C" int srandom(int);

/* Globals */

char alphas[62];
static ACG gen1; 
char randomStrings[100][9];
int Seed;

int numDigits(unsigned int num)
{
int nDigits = 0;

  while (num=num/10)
    nDigits++;
  nDigits++;
  return nDigits;
}

void print(FILE *fptr, unsigned int x, unsigned int y, char *str)
{
int nspaces;
char space = ' ';

  fprintf(fptr,"%ld",x);
  nspaces = 16 - numDigits(x);
  while (nspaces--)
   fprintf(fptr,"%c",space);
  fprintf(fptr,"%ld",y);
  nspaces = 16 - numDigits(y);
  while (nspaces--)
   fprintf(fptr,"%c",space);
  fprintf(fptr,"%s\n",str);    
}

void string(char *str)
{
int i = gen1.asLong()%100;

  Assert(str);
  strcpy(str,randomStrings[i]);
}

void InitString()
{
int i;
char str[8];

 for (int j=0;j<100;j++)
 {
   i = gen1.asLong()%62;
   str[0] = alphas[i];

   i = gen1.asLong()%62;
   str[1] = alphas[i];

   i = gen1.asLong()%62;
   str[2] = alphas[i];

   i = gen1.asLong()%62;
   str[3] = alphas[i];

   i = gen1.asLong()%62;
   str[4] = alphas[i];

  i = gen1.asLong()%62;
   str[5] = alphas[i];

  i = gen1.asLong()%62;
  str[6] = alphas[i];

  i = gen1.asLong()%62;
  str[7] = alphas[i];

  str[8] = '\0';
  strcpy(randomStrings[j],str);
 }

}

/* Creates all points in the specified extent */

void InsertAll(FILE *fptr, unsigned step)
{
char str[9];

  for (unsigned i = 0; i < extentp2x; i+=step)
  {
    for (unsigned j = 0; j < extentp2y; j+=step)
    {
      string(str);
      print(fptr,i,j,str);
    }
  }
}

/* Creates only diagonal points in the specified extent */

void InsertDiagonal(FILE *fptr, int step)
{
char str[9];

  for (int i = 0; i < extentp2x; i+=step)
  {
    string(str);
    print(fptr,i,i,str);
  }
}

/* Creates Diagonal points in random in the specified extent */

void InsertRandomDiagonal(FILE *fptr, int count)
{
char str[9];
ACG *gen1 = new ACG(Seed); 
int interval = 0;          /* When to change seed next */
int numberofchanges = 0;   /* What will be the seed */
int newseed;

  while (count--)
  {
    int x = gen1->asLong()%extentp2x;
    string(str);
    print(fptr, x, x, str);

    interval++;
    /* Fix seed for next change */
    if (interval == numberofchanges)
    {
      newseed = x;
      numberofchanges++;
    }
   /* time to change the random number generator ? */
    if (interval == 3000)
    {
      delete gen1;
      gen1 = new ACG(newseed);
      interval = 0;
    }
  }
}     

void InsertRandom(FILE *fptr, int count)
{
char str[9];
ACG *gen1 = new ACG(Seed); 
int interval = 0;          /* When to change seed next */
int numberofchanges = 0;   /* What will be the seed */
int newseed;
int toggle = 0;

  while (count--)
  {
    Point_t pt(gen1->asLong()%extentp2x, gen1->asLong()%extentp2y);
    string(str);
    print(fptr, pt.x, pt.y, str);
    
    interval++;
    /* Fix seed for next change */
    if (interval == numberofchanges)
    {
      newseed = (toggle == 0)? pt.x : pt.y;
      toggle = (toggle == 0)? 1: 0;
      numberofchanges++;
    }
    /* time to change the random number generator ? */
    if (interval == 3000)
    {
      delete gen1;
      gen1 = new ACG(newseed);
      interval = 0;
    }
  }
}


void InsertSkewed(FILE *fptr, int count, Point_t corner[], int size)
{
Point_t pt;
char str[9];
ACG *gen1 = new ACG(Seed); 
int interval = 0;          /* When to change seed next */
int numberofchanges = 0;   /* What will be the seed */
int newseed;
int toggle = 0;

  while (count--)
  {
    int j = gen1->asLong()%(size+1);
    if (j == size)                      // put on background
      pt = Point_t(gen1->asLong()%extentp2x, gen1->asLong()%extentp2y);
    else
    {
      pt.x = corner[j].x +  gen1->asLong()%influence;
      pt.y = corner[j].y +  gen1->asLong()%influence;
    }
    string(str);
    print(fptr,pt.x, pt.y,str);
   
     interval++;
    /* Fix seed for next change */
    if (interval == numberofchanges)
    {
      newseed = (toggle == 0)? pt.x : pt.y;
      toggle = (toggle == 0)? 1: 0;
      numberofchanges++;
    }
   /* time to change the random number generator ? */
    if (interval == 3000)
    {
      delete gen1;
      gen1 = new ACG(newseed);
      interval = 0;
    }
  }
}

void usage()
{
  fprintf(stderr,"Usage: generate <num points> <Random Number Seed>
                 <type:All or Diag or RandDiag or Rand or Skewed>
                 <outputfile>\n");
  exit(-1);
}

main(int argc, char **argv)
{
  if (argc < 5)
   usage();
  
  /* Check arguments */
  Assert(argv[1]);
  Assert(argv[2]);
  Assert(argv[3]);
  Assert(argv[4]);

  /* Open output file */
FILE *fptr = fopen(argv[4],"w");
  if (fptr == NULL)
  {
    fprintf(stderr,"Unable to open %s\n", argv[4]);
    exit(-1);
  }

int nPoints = atoi(argv[1]);
    Seed = atoi(argv[2]);
  /* Initialize the alphabets array used to generate random strings */

int i;
  for (i=0; i<=9; alphas[i]=i+'0',i++);
  for (char a='A'; a<='Z'; alphas[i]=a,i++,a++);
  for (a='a'; a<='z'; alphas[i]=a,i++,a++);
  InitString();

  /* What type to generate ? */

  if (!strcmp(argv[3],"All"))
    InsertAll(fptr, 20000);
  else if (!strcmp(argv[3],"Diag"))
    InsertDiagonal(fptr,step);
  else if (!strcmp(argv[3],"RandDiag"))
    InsertRandomDiagonal(fptr,nPoints);
  else if (!strcmp(argv[3],"Rand"))
    InsertRandom(fptr,nPoints);
  else if (!strcmp(argv[3],"Skewed"))
  {
    Point_t corner[NCorners];
    for (int i = 0; i < NCorners; i++)
      corner[i] = Point_t(i*65536, i*65536);
    InsertSkewed(fptr,nPoints,corner,NCorners);
  }
  else
   Assert(0);    
  fclose(fptr);
}






