#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <iostream.h>
#include <fcntl.h>
#include <errno.h>
#include <Random.h>

#include "standard.h"


struct Point_t
{
  unsigned int x;
  unsigned int y;
  Point_t(unsigned int xx=0, unsigned int yy=0) :x(xx), y(yy) {}

  void Print()
  {
	cout << '(' << x << ',' << y << ')' << endl ;
  }	

  operator == (const Point_t& aPoint)
  {
	return x == aPoint.x && y == aPoint.y ;
  }	

  operator < (const Point_t& aPoint)
  {
	return x < aPoint.x && y < aPoint.y ;
  }

  friend ostream&
  operator << (ostream& os, const Point_t& aPoint)
  {
	cout << '(' << aPoint.x << ',' << aPoint.y << ')' ;

	return os ; 
  }
};


/* Some constants used in data generation */

const unsigned extentp2x = 1<<24;
const unsigned extentp2y = 1<<24;
const int influence = 1024;
const int step = 2;
const int NCorners = 32;


extern "C" long random(void);
extern "C" int srandom(int);

/* Globals */

char alphas[62];
int randomSeeds[50];
const int NumSeeds = 50;
int interval;

int numDigits(unsigned int num)
{
int nDigits = 0;

  while (num=num/10)
    nDigits++;
  nDigits++;
  return nDigits;
}

void print(FILE *fptr, unsigned int x, unsigned int y, char *str)
{
int nspaces;
char space = ' ';

  fprintf(fptr,"%ld",x);
  nspaces = 16 - numDigits(x);
  while (nspaces--)
   fprintf(fptr,"%c",space);
  fprintf(fptr,"%ld",y);
  nspaces = 16 - numDigits(y);
  while (nspaces--)
   fprintf(fptr,"%c",space);
  fprintf(fptr,"%s\n",str);    
}

void string(char *str,int seed)
{
int i;

  srandom(seed);
  i = random()%62;
  str[0] = alphas[i];

  i = random()%62;
  str[1] = alphas[i];

  i = random()%62;
  str[2] = alphas[i];

  i = random()%62;
  str[3] = alphas[i];

  i = random()%62;
  str[4] = alphas[i];

  i = random()%62;
  str[5] = alphas[i];

  i = random()%62;
  str[6] = alphas[i];

  i = random()%62;
  str[7] = alphas[i];

  str[8] = '\0';
}

/* Creates all points in the specified extent */

void InsertAll(FILE *fptr, int step, Random *gen)
{
char str[9];

  for (int i = 0; i < extentp2x; i+=step)
  {
    for (int j = 0; j < extentp2y; j+=step)
    {
      string(str,i*j);
      print(fptr,i,j,str);
    }
  }
}

/* Creates only diagonal points in the specified extent */

void InsertDiagonal(FILE *fptr, int step, Random *gen)
{
char str[9];

  for (int i = 0; i < extentp2x; i+=step)
  {
    string(str,i);
    print(fptr,i,i,str);
  }
}

/* Creates Diagonal points in random in the specified extent */

void InsertRandomDiagonal(FILE *fptr, int count, Random *gen)
{
int i = 0;
char str[9];
int n, intervalNum;

  n = 0;
  intervalNum = 0;
  while (i < count)
  {
    int x = random()%extentp2x;
    string(str,x);
    print(fptr, x, x, str);
    i += 1;
    n++;
    if (n==interval)
    { /* Change the random number seed at every 'interval' th generation */
      srandom(randomSeeds[intervalNum++]);
      n = 0;
    }
  }
}     

void InsertRandom(FILE *fptr, int count, Random *gen)
{
int i = 0;
char str[9];
int n, intervalNum;

  n = 0;
  intervalNum = 0;
  while (i < count)
  {
    Point_t pt(random()%extentp2x, random()%extentp2y);
    string(str,pt.x);
    print(fptr,pt.x, pt.y,str);
    i += 1;
    n++;    
    if (n==interval)
    {  /* Change the random number seed at every 'interval' th generation */
       srandom(randomSeeds[intervalNum++]);
       n = 0;
    }
  }
}


void InsertSkewed(FILE *fptr,int count,Point_t corner[],int size,Random *gen)
{
int i = 0;
Point_t pt;
char str[9];
int n, intervalNum;

  n = 0;
  intervalNum = 0;
  while (i < count)
  {
    int j = random() % (size+1);
    if (j == size)                      // put on background
      pt = Point_t(random()%extentp2x, random()%extentp2y);
    else
    {
      pt.x = corner[j].x + random()%influence;
      pt.y = corner[j].y + random()%influence;
    }
    string(str,pt.x);
    print(fptr,pt.x, pt.y,str);
    i += 1;
    n++;
    if (n==interval)
    {  /* Change the random number seed at every 'interval' th generation */
       srandom(randomSeeds[intervalNum++]);
       n = 0;
    }
  }
}

void usage()
{
  fprintf(stderr,"Usage: generate <num points> <Random Number Seed>
                 <type:All or Diag or RandDiag or Rand or Skewed>
                 <outputfile>\n");
  exit(-1);
}

main(int argc, char **argv)
{
  if (argc < 5)
   usage();
  
  /* Check arguments */
  Assert(argv[1]);
  Assert(argv[2]);
  Assert(argv[3]);
  Assert(argv[4]);

  /* Initialize random number generator */
Random generator(atoi(argv[2]));

  /* Open output file */
FILE *fptr = fopen(argv[4],"w");
  if (fptr == NULL)
  {
    fprintf(stderr,"Unable to open %s\n", argv[4]);
    exit(-1);
  }

int nPoints = atoi(argv[1]);

  interval = nPoints/NumSeeds;/* Interval when random number seed is changed */

  /* Initialize the alphabets array used to generate random strings */

int i;
  for (i=0; i<=9; alphas[i]=i+'0',i++);
  for (char a='A'; a<='Z'; alphas[i]=a,i++,a++);
  for (a='a'; a<='z'; alphas[i]=a,i++,a++);

  /* Initialise the random number seeds */
  for (i=0;i<NumSeeds;randomSeeds[i]=random(),i++);

  /* What type to generate ? */

  if (!strcmp(argv[3],"All"))
    InsertAll(fptr, step, &generator);
  else if (!strcmp(argv[3],"Diag"))
    InsertDiagonal(fptr,step,&generator);
  else if (!strcmp(argv[3],"RandDiag"))
    InsertRandomDiagonal(fptr,nPoints,&generator);
  else if (!strcmp(argv[3],"Rand"))
    InsertRandom(fptr,nPoints,&generator);
  else if (!strcmp(argv[3],"Skewed"))
  {
    Point_t corner[NCorners];
    for (int i = 0; i < NCorners; i++)
      corner[i] = Point_t(i*65536, i*65536);
    InsertSkewed(fptr,nPoints,corner,NCorners,&generator);
  }
  else
   Assert(0);    
  fclose(fptr);
}






