#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <iostream.h>
#include <math.h>
#include <fcntl.h>

#include <standard.h>
#include <qtree.h>
#include <quadsort.h>

int mycompare(const void *p1, const void *p2)
{
	Point_t pp1,pp2;
 
	pp1.x = ((Point_t*)p1)->x;
	pp1.y = ((Point_t*)p1)->y;
	pp2.x = ((Point_t*)p2)->x;
	pp2.y = ((Point_t*)p2)->y;

	if (compare(pp1, pp2) && compare(pp2, pp1))
		return 0;

	if (!compare(pp1, pp2))
		return -1;

	return 1;
}

// --------------------------------------------------------------------------
// returns 1 if morton(p1) >= morton(p2), 0 otherwise 
// --------------------------------------------------------------------------
int compare(Point_t p1, Point_t p2)
{
	unsigned int dx = (unsigned)p1.x ^ (unsigned)p2.x;
	unsigned int dy = (unsigned)p1.y ^ (unsigned)p2.y;
	unsigned int d  = dx ^ dy;

	return ((d<dx && d<dy) || dy>dx) ? p1.y >= p2.y : p1.x >= p2.x;
}

Quadfile_t::Quadfile_t(char *fName)
{
	int nbytes;
  
	points[0].x = points[0].y = 0 ;
	points[1].x = points[1].y = 0 ;

	curIndex = 0 ;

  	fDesc = open(fName,O_RDONLY,0);
  	if (fDesc < 0)
  	{
    		fprintf(stderr,"Error opening sorted file: ");
    		perror("");
    		exit(-1);
	}

  	if ( (nbytes = read(fDesc,(char*)&points[curr()],sizeof(Point_t)))
       		!= sizeof(Point_t) )
  	{
		setNonValid(curr()) ;
		setNonValid(next()) ; 
		return ;
  	}

	setValid( curr() ) ;
	
	if ( (nbytes = read(fDesc,(char*)&points[next()],sizeof(Point_t)))
       		!= sizeof(Point_t) )
	{
		setNonValid(next()) ;
		return ;
	}

	setValid( next() ) ;
}

Quadfile_t::~Quadfile_t()
{
	close(fDesc);
}

int 
Quadfile_t::peek(Point_t pt)
{
	if ( !extent.Contains(pt) )	 	// Invalid pt , cannot peek 
		return -1;			

	if ( !isValid( curr() ) )		// possible end of file
	{
		Assert( !isValid( next() ) );	// second better be invalid too 
		return -1;
	}

	if ( !compare(pt, points[curr()]) )  	// pt < first, return 0 
		return NO_POINT ;

	if ( compare( pt, points[curr()] ) && compare(points[curr()], pt) )    // pt == first,return 1 
		return SINGLE_POINT;

	// pt > first is true 
	Assert(compare(pt, points[curr()]) && !compare(points[curr()], pt));

	if ( !isValid( next() ) )
		return SINGLE_POINT;

	if ( !compare( pt, points[next()]) )          // pt < second 
		return SINGLE_POINT;

	if ( compare(pt,points[next()]) && compare(points[next()],pt) ) // pt == second 
		return MANY_POINTS;

	return MANY_POINTS;            // pt > second
}

int 
Quadfile_t::Get( Point_t& pt )
{
	int nbytes ;

	if ( !isValid( curr() ) )  
		return -1 ; 

	// Get the point at the curr index
	pt.x = points[curr()].x ;
	pt.y = points[curr()].y ;

	// Get a new point from the file
  	if ( (nbytes = read(fDesc,(char*)&points[curr()],sizeof(Point_t)))
       		!= sizeof(Point_t) )
  	{
		setNonValid( curr() ) ;
  	}

	// Go to the next index 
	incr() ;

	return 0 ; 
}

void usage()
{
  fprintf(stderr,"Usage: quadsort <infile> \n");
  exit(-1);
}

/*
main(int argc, char *argv[])
{
	if (argc < 2)
		usage();

	Quadfile_t file(argv[1]) ;

	file.setExtent( Rect_t(xWorldMin,yWorldMin,xWorldMax,yWorldMax) ) ;

	Point_t	pt ;

	while ( file.Get(pt) >= 0 )
	{
		pt.Print() ;
	}
}
*/
